@@ -24,17 +24,29 @@ __all__ = [
     "read_config_file_dict",
 ]
 try:
-    import tomllib
-except Exception:
+    import tomllib  # novermin
+except Exception:  # noqa: BLE001
     import tomli as tomllib
 
 import contextlib
 import os
 import pathlib
+import shlex
 import subprocess  # noqa: S404
 import sys
-from typing import Any, Generator, TypeVar, cast
-import shlex
+from typing import (
+    Any,
+    Dict,
+    Generator,
+    Iterable,
+    List,
+    Mapping,
+    MutableSequence,
+    NamedTuple,
+    Optional,
+    TypeVar,
+    cast,
+)
 
 import pydantic
 from typing_extensions import Self
@@ -47,18 +59,18 @@ CONFIG_FILE_NOT_FOUND_EXC = FileNotFoundError(
     " file, or an [tool.lemming] entry in the pyproject.toml file!"
 )
 POTENTIAL_ERRORS = {
-    "No module named".casefold(): "HINT: It seems like that module doesn't"
+    b"No module named".lower(): "HINT: It seems like that module doesn't"
     " exist with that name. If the module has to be ran with a separate name"
     " than the one that was used to install it, add the `install_name` option"
     " to the config.",
-    "Permission denied: '.'".casefold(): "HINT: It seems like that the"
+    b"Permission denied: '.'".lower(): "HINT: It seems like that the"
     " formatter/linter expects a file as an argument, and not a directory."
     " Lemming by default doesn't pass anything to the formatter/linter as"
     " arguments. You can add arguments with the `args` config key. You can"
     " also use `{path}` in the `args` config key, which will be replaced with"
     " the path that was passed to Lemming, or the current working directory by"
     " default.",
-    "Usage:".casefold(): "HINT: It seems like that the formatter/linter"
+    b"Usage:".lower(): "HINT: It seems like that the formatter/linter"
     " expects arguments. You can pass the `args` config key to do that. Note,"
     " that `{path}` in the `args` config key will be replaced by the path that"
     " was passed to Lemming, or the current working directory by default.",
@@ -66,13 +78,20 @@ POTENTIAL_ERRORS = {
 T = TypeVar("T")
 
 
-def assert_dict_keys(dictionary: dict[Any, Any], keys: list[Any]) -> None:
+class WhatToQuiet(NamedTuple):
+    commands: bool
+    pip: bool
+
+
+def assert_dict_keys(
+    dictionary: Iterable[T], keys: MutableSequence[T]
+) -> None:
     """
     Assert that `dictionary`'s keys are in `keys`
 
     Args:
-        dictionary (dict[Any, Any]): The dictionary to check.
-        keys (list[Any]): The keys that are allowed.
+        dictionary (Iterable[Any, Any]): The dictionary to check.
+        keys (MutableSequence[Any]): The keys that are allowed.
 
     Raises:
         ValueError: If a key is found within `dictionary` but not within
@@ -92,7 +111,7 @@ def assert_dict_keys(dictionary: dict[Any, Any], keys: list[Any]) -> None:
             keys.remove(key)
 
 
-def read_toml_file(file: pathlib.Path) -> dict[str, Any]:
+def read_toml_file(file: pathlib.Path) -> Dict[str, Any]:
     """
     Read the toml file `file`.
 
@@ -104,7 +123,7 @@ def read_toml_file(file: pathlib.Path) -> dict[str, Any]:
     """
     logger.debug(f"Reading toml file {file}")
     with file.open("rb") as fd:
-        return cast(dict[str, Any], tomllib.load(fd))  # type: ignore
+        return cast(Dict[str, Any], tomllib.load(fd))  # type: ignore
 
 
 def get_home() -> pathlib.Path:
@@ -118,7 +137,7 @@ def get_home() -> pathlib.Path:
     return pathlib.Path.home()
 
 
-def get_xdg_config_home() -> pathlib.Path | None:
+def get_xdg_config_home() -> Optional[pathlib.Path]:
     """
     Get $XDG_CONFIG_HOME or the default value or None. Return value will
     always be an existing directory (unless None).
@@ -168,7 +187,7 @@ def _get_xdg_config_dirs() -> Generator[pathlib.Path, None, None]:
                     yield path
 
 
-def get_xdg_config_dirs() -> list[pathlib.Path]:
+def get_xdg_config_dirs() -> List[pathlib.Path]:
     """
     Get $XDG_CONFIG_DIRS. Return value will always be a list of existing
     directories.
@@ -275,36 +294,14 @@ def find_config_file(
 class FormatterOrLinter(pydantic.BaseModel):
     """
     An "ABC" for formatters and linters.
-    Does all of the work except the method .run().
 
     Args:
-        package (str): The package's name. This will be used to install and
-        run it!
-        version (str | None, optional): The package's version. If None it will
-        be the latest stable version. Defaults to None.
-        args (str, optional): Arguments to the formatter/linter. Defaults
-        to "".
-        allow_nonzero (bool, optional): Allow non-zero exit status for
-        formatters (not linters!). Defaults to False.
-        install_name (str | None, optional): The package's name to install. If
-        it's None, it defaults to `package`. Defaults to None.
+        packages (List[str]): The packages' name (optionally versions
+        with "==x.y.z") to install with pip.
     """
 
-    package: str
-    version: str | None = None
-    args: str = ""
-    allow_nonzero: bool = False
-    also_install: list[str] = []
-    install_name: str | None = None
-
-    def get_version(self) -> str:
-        """
-        Get the version stuff.
-
-        Returns:
-            str: f"=={self.version}" if self.version else ""
-        """
-        return f"=={self.version}" if self.version else ""
+    # it's actually Iterable, but that introduces some bugs
+    packages: List[str]
 
     @staticmethod
     def get_pyexe() -> str:
@@ -319,22 +316,19 @@ class FormatterOrLinter(pydantic.BaseModel):
     def run_command(
         self,
         cmd: str,
-        paths_to_check: list[pathlib.Path],
+        paths_to_check: Iterable[pathlib.Path],
         quiet: bool,
     ) -> bool:
         """
         Run command `cmd`.
         Also replaces these:
         - `{pyexe}` -> `sys.executable`
-        - `{package}` -> `self.package`
-        - `{also_install}` -> `self.also_install`
-        - `{args}` -> `self.args`
 
         In `self.args` `{path}` be replaced by `paths_to_check`.
 
         Args:
             cmd (str): The command to run.
-            paths_to_check (list[pathlib.Path]): The paths to check.
+            paths_to_check (Iterable[pathlib.Path]): The paths to check.
             quiet (bool): Don't let the command write to stdout and stderr
 
         Returns:
@@ -342,12 +336,10 @@ class FormatterOrLinter(pydantic.BaseModel):
         """
         cmd = cmd.strip()
         cmd = cmd.replace("{pyexe}", sys.executable)
-        cmd = cmd.replace("{package}", self.package)
-        cmd = cmd.replace("{args}", self.args)
         paths = " ".join(map(str, paths_to_check))
         cmd = cmd.replace("{path}", paths)
-        also_install = " ".join(self.also_install)
-        cmd = cmd.replace("{also_install}", also_install)
+        packages = " ".join(self.packages)
+        cmd = cmd.replace("{packages}", packages)
 
         splitted = shlex.split(cmd, posix=os.name != "nt")
         logger.debug(f"Running command {splitted!r}")
@@ -363,12 +355,16 @@ class FormatterOrLinter(pydantic.BaseModel):
         logger.error(
             f"Command {cmd!r} returned non-zero exit status {exit_status}"
         )
+        if completed_process.stdout:
+            stdout = completed_process.stderr.lower()
+            for look_for, suggestion in POTENTIAL_ERRORS.items():
+                if look_for in stdout:
+                    logger.error(suggestion)
         return False
 
     def install(self, quiet: bool) -> bool:
         """
-        Install.
-        Runs the following command: `{pyexe} -m pip install -U {package}`
+        Install the packages.
 
         Args:
             quiet (bool): Don't let `pip` write to stdout and stderr.
@@ -376,97 +372,118 @@ class FormatterOrLinter(pydantic.BaseModel):
         Returns:
             bool: `exit_status == 0`
         """
-        logger.info(f"Installing {self.package} (and {self.also_install})")
+        logger.info(f"Installing {self.packages}")
         with logger.ctxmgr:
             return self.run_command(
-                "{pyexe} -m pip install -U {package} {also_install}", [], quiet
+                "{pyexe} -m pip install -U {packages}", [], quiet
             )
 
-    def _run(self, paths_to_check: list[pathlib.Path], quiet: bool) -> bool:
-        return self.run_command(
-            "{pyexe} -m {package} {args}", paths_to_check, quiet
-        )
-
     def run(
         self,
-        paths_to_check: list[pathlib.Path],
-        what_to_quiet: tuple[bool, bool],
+        paths_to_check: Iterable[pathlib.Path],
+        what_to_quiet: WhatToQuiet,
     ) -> bool:
         raise NotImplementedError
 
-    @classmethod
-    def from_dict(cls, dictionary: dict[str, Any]) -> Self:
-        """
-        Return `Self` from a dictionary.
-
-        Args:
-            dictionary (dict[str, Any]): The dictionary to use.
-
-        Returns:
-            Self: The new object from `dictionary`
-        """
-        assert_dict_keys(
-            dictionary,
-            [
-                "package",
-                "version",
-                "args",
-                "allow_nonzero",
-                "also_install",
-                "install_name",
-            ],
-        )
-        return cls(**dictionary)
-
 
 class Formatter(FormatterOrLinter):
-    def run(
+    format_command: str
+    check_command: Optional[str] = None
+    allow_nonzero_on_format: bool = False
+
+    def run_format(
         self,
-        paths_to_check: list[pathlib.Path],
-        what_to_quiet: tuple[bool, bool],
+        paths_to_check: Iterable[pathlib.Path],
+        what_to_quiet: WhatToQuiet,
     ) -> bool:
-        if not self.install(what_to_quiet[1]):
+        if not self.install(what_to_quiet.pip):
             logger.error(
-                f"Could not install formatter {self.package}! See"
+                f"Could not install the packages {self.packages}! See"
                 " pip's output for more information."
             )
-        success = self._run(paths_to_check, what_to_quiet[0])
-        if success or self.allow_nonzero:
-            logger.info(f"Successfully ran formatter {self.package}!")
+            return False
+        success = self.run_command(
+            self.format_command, paths_to_check, what_to_quiet.commands
+        )
+        if success or self.allow_nonzero_on_format:
+            logger.info(
+                f"Successfully ran (format) formatter {self.packages}!"
+            )
         else:
             logger.error(
-                f"Could not run formatter {self.package} ({self.version = !r};"
-                f" {self.args = !r}; {paths_to_check = !r})! (NOTE: Formatters"
-                " are expected to MODIFY the code if the formatter finds a"
-                " violation (not just simply check the code). Formatters are"
-                " also expected to return a ZERO exit status even if the"
-                " formatter had to modify a file. To allow this behavior"
-                ' change "allow_nonzero" to true in the config file for this'
-                " formatter.)"
+                f"Could not run (format) formatter ({self.packages};"
+                f" {self.format_command!r})! (NOTE: The format_command command"
+                " is expected to modify/format the code and return zero. This"
+                " is DIFFERENT from check_command. If you still want to allow"
+                " this, set allow_nonzero_on_format=true for this formatter.)"
             )
         return success
 
-
-class Linter(FormatterOrLinter):
-    def run(
+    def run_check(
         self,
-        paths_to_check: list[pathlib.Path],
-        what_to_quiet: tuple[bool, bool],
+        paths_to_check: Iterable[pathlib.Path],
+        what_to_quiet: WhatToQuiet,
     ) -> bool:
-        if not self.install(what_to_quiet[1]):
+        if not self.check_command:
+            logger.warning(
+                f"The formatter {self.packages} does NOT have a check_command!"
+                " Skipping..."
+            )
+            return True
+        if not self.install(what_to_quiet.pip):
             logger.error(
-                f"Could not install linter {self.package}! See"
+                f"Could not install the packages {self.packages}! See"
                 " pip's output for more information."
             )
-        if self.allow_nonzero:
-            logger.warning(
-                '"allow_nonzero" is set to true in the config file! This'
-                " attribute is only supported for formatters, not linters!"
-                " Ignoring"
-            )
-        success = self._run(paths_to_check, what_to_quiet[0])
+            return False
+        success = self.run_command(
+            self.check_command, paths_to_check, what_to_quiet.commands
+        )
         if success:
-            logger.info(f"Successfully ran linter {self.package}!")
+            logger.info(f"Successfully ran (check) formatter {self.packages}!")
+        else:
+            logger.error(
+                f"Could not run (check) formatter ({self.packages};"
+                f" {self.format_command!r})! (NOTE: The check_command command"
+                " is expected to CHECK that the code is up to standards. If"
+                " the code is ok, the command should return 0; otherwise it"
+                " should return non-zero.)"
+            )
+        return success
+
+    @classmethod
+    def from_dict(cls, dict_: Mapping[str, Any]) -> Self:
+        assert_dict_keys(
+            dict_,
+            [
+                "packages",
+                "format_command",
+                "check_command",
+                "allow_nonzero_on_format",
+            ],
+        )
+        return cls(**dict_)
+
+
+class Linter(FormatterOrLinter):
+    command: str
+    run_first: bool = False
+
+    def run(
+        self,
+        paths_to_check: Iterable[pathlib.Path],
+        what_to_quiet: WhatToQuiet,
+    ) -> bool:
+        if not self.install(what_to_quiet.pip):
+            logger.error(
+                f"Could not install linter {self.packages}! See"
+                " pip's output for more information."
+            )
+        success = self.run_command(
+            self.command, paths_to_check, what_to_quiet.commands
+        )
+        if success:
+            logger.info(f"Successfully ran linter {self.packages}!")
         else:
             logger.error(
                 "Linting failed! See the linter's output for more"
@@ -474,16 +491,30 @@ class Linter(FormatterOrLinter):
             )
         return success
 
+    @classmethod
+    def from_dict(cls, dict_: Mapping[str, Any]) -> Self:
+        assert_dict_keys(
+            dict_,
+            ["packages", "command", "run_first"],
+        )
+        return cls(**dict_)
+
 
 class Config(pydantic.BaseModel):
-    formatters: list[Formatter]
-    linters: list[Linter]
+    formatters: Iterable[Formatter]
+    linters: Iterable[Linter]
     # ^ also modify within read_config_file()
 
+    def get_first_linters(self) -> List[Linter]:
+        return [linter for linter in self.linters if linter.run_first]
+
+    def get_other_linters(self) -> List[Linter]:
+        return [linter for linter in self.linters if not linter.run_first]
+
 
 def read_config_file_dict(
-    read_data: dict[str, Any], file: str = "<unknown file>"
-) -> dict[str, Any]:
+    read_data: Mapping[str, Any], file: str = "<unknown file>"
+) -> Dict[str, Any]:
     try:
         logger.debug('Searching for "lemming" key')
         data = read_data["lemming"]
@@ -510,7 +541,7 @@ def read_config_file_dict(
     return data
 
 
-def read_config_file(dictionary: dict[str, Any]) -> Config:
+def read_config_file(dictionary: Mapping[str, Any]) -> Config:
     logger.debug(f"Making dict {dictionary} a Config() object")
     with logger.ctxmgr:
         assert_dict_keys(dictionary, ["formatters", "linters"])
